use std::collections::HashMap;
use std::io;
use std::net::{SocketAddr, TcpListener};
use std::os::fd::{AsRawFd, RawFd};

use crate::epoll::{Epoll, is_read_event, is_write_event};
use crate::RequestContext;

const KEY_NEW_CONNECTION: u64 = 100;

pub struct TcpServer {
    addr: SocketAddr,
    listener: TcpListener,
    listener_fd: RawFd,
    epoll: Epoll,
    request_contexts: HashMap<u64, RequestContext>,
    key: u64,
}

impl TcpServer {
    pub fn new(addr: SocketAddr) -> io::Result<Self> {
        let listener = TcpListener::bind(addr)?;
        listener.set_nonblocking(true)?;

        let listener_fd = listener.as_raw_fd();

        let epoll = Epoll::create()?;
        epoll.add_read_interest(listener_fd, KEY_NEW_CONNECTION)?;

        Ok(TcpServer {
            addr,
            listener,
            listener_fd,
            epoll,
            request_contexts: HashMap::new(),
            key: KEY_NEW_CONNECTION,
        })
    }

    pub fn listen(&mut self) -> io::Result<()> {
        println!("Listening on {}", self.addr);

        loop {
            self.epoll.wait().expect("Failed to wait for epoll event");

            let events = &self.epoll.events.iter()
                .map(|event| (event.events, event.u64))
                .collect::<Vec<(u32, u64)>>();

            for (events, u64) in events {
                match *u64 {
                    KEY_NEW_CONNECTION => self.accept_connection()?,
                    key => self.handle_event(*events, key)
                }
            }
        }
    }

    fn accept_connection(&mut self) -> io::Result<()> {
        match self.listener.accept() {
            Ok((stream, addr)) => {
                stream.set_nonblocking(true)?;
                println!("New client: {addr}");
                self.key += 1;
                self.epoll.add_read_interest(stream.as_raw_fd(), self.key)?;
                self.request_contexts.insert(self.key, RequestContext::new(stream));
            }
            Err(e) => eprintln!("Couldn't accept: {e}")
        };

        self.epoll.modify_read_interest(self.listener_fd, KEY_NEW_CONNECTION)
    }

    fn handle_event(&mut self, events: u32, key: u64) {
        let mut to_delete = None;
        if let Some(context) = self.request_contexts.get_mut(&key) {
            match events {
                v if is_read_event(v) => {
                    // context.read_cb(key, epoll_fd)?;
                }
                v if is_write_event(v) => {
                    // context.write_cb(key, epoll_fd)?;
                    to_delete = Some(key);
                }
                v => println!("Unexpected event: {v}"),
            }
        }
        if let Some(key) = to_delete {
            self.request_contexts.remove(&key);
        }
    }
}
